import abc
import inspect

from hnhm.core import Group, Entity, Layout, HnhmError, ChangeType, LayoutType

from .hnhm_attribute import HnhmAttribute


class HnhmEntity(abc.ABC):
    __layout__: Layout = None
    __keys__: list[HnhmAttribute] | None = None

    def to_core(self) -> Entity:
        inspected = dict(inspect.getmembers(self))

        if not inspected.get("__layout__"):
            raise HnhmError(
                f"Layout not found for entity: '{self}'."
                " Please, specify Layout via '__layout__' attribute."
            )
        layout: Layout = inspected["__layout__"]
        if not layout.type:
            raise HnhmError(
                f"Type for Layout '{layout}' is required for entity."
                " Please, specify LayoutType via 'type' attribute."
            )

        name = layout.name

        if not inspected.get("__doc__"):
            raise HnhmError(
                f"Doc not found or empty for entity: '{layout.type}.{name}'."
                " Please, write a documentation for your entity."
            )
        doc: str = inspected["__doc__"]

        match layout.type:
            case LayoutType.STAGE:
                keys = []

            case LayoutType.HNHM:
                if not inspected.get("__keys__"):
                    raise HnhmError(
                        f"At least one Key is required for entity '{layout.type}.{name}'."
                        " Please, specify entity's keys via the '__keys__' attribute."
                    )

                keys_hnhm: list[HnhmAttribute] = inspected["__keys__"]
                for key in keys_hnhm:
                    if key.change_type != ChangeType.IGNORE:
                        raise HnhmError(
                            f"Change type='{key.change_type}' is not supported for Key attributes."
                            f" Use 'ChangeType.IGNORE' for the key attributes in the '{layout.type}.{name}' entity."
                        )

                keys = [key.to_core(name) for key in keys_hnhm]
                if len(keys) != len(set(keys)):
                    raise HnhmError(
                        f"Found duplicated keys for entity: '{layout.type}.{name}'."
                    )

            case _:
                raise HnhmError(f"Unknown LayoutType='{layout.type}'")

        groups = {}
        attributes = {}
        for class_attribute in inspected.values():
            if not isinstance(class_attribute, HnhmAttribute):
                continue

            attribute = class_attribute.to_core(name)
            attribute_name = attribute.name

            if attribute.group:
                group_name = attribute.group
                if group_name not in groups:
                    groups[group_name] = Group(
                        name=group_name,
                        entity_name=name,
                        attributes={attribute_name: attribute},
                        change_type=attribute.change_type,
                    )

                if groups[group_name].change_type != attribute.change_type:
                    raise HnhmError(
                        f"Found conflicting ChangeType for the entity='{layout.type}.{name}' group='{group_name}'."
                        " Please, use single ChangeType for all attributes within the same group."
                    )
                groups[group_name].attributes[attribute_name] = attribute

            elif attribute not in keys:
                attributes[attribute_name] = attribute

        if layout.type == LayoutType.STAGE and len(attributes) < 1:
            raise HnhmError(
                f"Entity='{layout.type}.{name}' should have at least 1 attribute."
            )

        return Entity(
            name=name,
            layout=layout,
            doc=doc,
            keys=keys,
            attributes=attributes,
            groups=groups,
        )



from enum import Enum

import pydantic

from hnhm.core import (
    Sql,
    Entity,
    Storage,
    HnhmError,
    Migration,
    CreateLink,
    LayoutType,
    RemoveLink,
    CreateGroup,
    RemoveGroup,
    CreateEntity,
    RemoveEntity,
    CreateAttribute,
    RemoveAttribute,
    RemoveEntityView,
    AddGroupAttribute,
    RecreateEntityView,
    RemoveGroupAttribute,
)

from .hnhm_link import HnhmLink
from .hnhm_entity import HnhmEntity


class PlanType(str, Enum):
    CREATE = "CREATE"
    REMOVE = "REMOVE"
    UPDATE = "UPDATE"


class PlanCollection(pydantic.BaseModel):
    type: PlanType
    migrations: list[Migration]


class Plan(pydantic.BaseModel):
    entities_migrations: dict[str, PlanCollection]
    links_migrations: dict[str, PlanCollection]

    def is_empty(self):
        return not self.entities_migrations and not self.links_migrations

    @property
    def migrations_all(self):
        migrations: list[Migration] = []
        for collection in self.entities_migrations.values():
            migrations.extend(collection.migrations)
        for collection in self.links_migrations.values():
            migrations.extend(collection.migrations)
        migrations = sorted(migrations, key=lambda m: m.priority)
        return migrations


class HnHm:
    def __init__(self, *, sql: Sql, storage: Storage):
        self.sql = sql
        self.storage = storage
        self.data = self.storage.load()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.storage.save(self.data)

    def plan(
        self,
        *,
        entities: list[HnhmEntity] | None = None,
        links: list[HnhmLink] | None = None,
    ) -> Plan:
        core_entities = {}
        for entity in entities or []:
            entity_core = entity.to_core()
            core_entities[entity_core.name] = entity_core

        core_links = {}
        for link in links or []:
            link_core = link.to_core()
            core_links[link_core.name] = link_core

        plan = Plan(entities_migrations={}, links_migrations={})

        # Entity: create if not exists
        for entity in core_entities.values():
            migrations = []

            # Entity's View: create/update if not exists
            if (
                entity.name not in self.data.entities_views
                and entity.layout.type == LayoutType.HNHM
            ):
                migrations.append(RecreateEntityView(entity=entity))

            # Create Entity
            if entity.name not in self.data.entities:
                migrations.append(CreateEntity(entity=entity))

                if entity.layout.type == LayoutType.HNHM:
                    # Create Attribute
                    for attribute in entity.attributes.values():
                        migrations.append(
                            CreateAttribute(entity=entity, attribute=attribute)
                        )
                    # Create Group
                    for group in entity.groups.values():
                        migrations.append(CreateGroup(entity=entity, group=group))

            if migrations:
                plan.entities_migrations[entity.name] = PlanCollection(
                    type=PlanType.CREATE,
                    migrations=migrations,
                )

        # Entity: create/remove/update Attribute/Group
        for entity in core_entities.values():
            if entity.name not in self.data.entities:
                continue

            attributes_state = self.data.entities[entity.name].attributes
            groups_state = self.data.entities[entity.name].groups

            migrations = []
            # Create Attribute
            for attribute_name, attribute in entity.attributes.items():
                if attribute_name not in attributes_state:
                    migrations.append(CreateAttribute(entity=entity, attribute=attribute))

            # Remove Attribute
            for attribute_name, attribute in attributes_state.items():
                if attribute_name not in entity.attributes:
                    migrations.append(RemoveAttribute(entity=entity, attribute=attribute))

            # Create/Update Group
            for group_name, group in entity.groups.items():
                # Update
                if group_name in groups_state:
                    group_state = groups_state[group_name]
                    # Add an Attribute to a Group
                    for attribute_name, attribute in group.attributes.items():
                        if attribute_name not in group_state.attributes:
                            migrations.append(
                                AddGroupAttribute(
                                    entity=entity, group=group, attribute=attribute
                                )
                            )
                    # Remove an Attribute from a Group
                    for attribute_name, attribute in group_state.attributes.items():
                        if attribute_name not in group.attributes:
                            migrations.append(
                                RemoveGroupAttribute(
                                    entity=entity, group=group, attribute=attribute
                                )
                            )
                # Create
                else:
                    migrations.append(CreateGroup(entity=entity, group=group))

            # Remove Group
            for group_name, group in groups_state.items():
                if group_name not in entity.groups:
                    migrations.append(RemoveGroup(entity=entity, group=group))

            if migrations:
                if entity.layout.type == LayoutType.HNHM:
                    migrations.extend(
                        [
                            RemoveEntityView(entity=entity),
                            RecreateEntityView(entity=entity),
                        ]
                    )
                plan.entities_migrations[entity.name] = PlanCollection(
                    type=PlanType.UPDATE,
                    migrations=migrations,
                )

        # Link: remove
        for link_name, link in self.data.links.items():
            if link_name not in core_links:
                plan.links_migrations[link_name] = PlanCollection(
                    type=PlanType.REMOVE,
                    migrations=[RemoveLink(link=link)],
                )

        # Entity: remove
        for entity_name, entity in self.data.entities.items():
            if entity_name not in core_entities:
                migrations = []

                if entity.layout.type == LayoutType.HNHM:
                    migrations.append(RemoveEntityView(entity=entity))

                    attributes_state = self.data.entities[entity_name].attributes
                    groups_state = self.data.entities[entity_name].groups
                    # Remove Attribute
                    for _, attribute_state in attributes_state.items():
                        migrations.append(
                            RemoveAttribute(entity=entity, attribute=attribute_state)
                        )
                    # Remove Group
                    for group_name, group in groups_state.items():
                        migrations.append(RemoveGroup(entity=entity, group=group))

                migrations.append(RemoveEntity(entity=entity))
                plan.entities_migrations[entity_name] = PlanCollection(
                    type=PlanType.REMOVE,
                    migrations=migrations,
                )

        # Link: create
        for link_name, link in core_links.items():
            if link_name not in self.data.links:
                plan.links_migrations[link.name] = PlanCollection(
                    type=PlanType.CREATE,
                    migrations=[CreateLink(link=link)],
                )

        return plan

    def apply(self, plan: Plan):
        for migration in plan.migrations_all:
            sql = self.sql.generate_sql(migration)

            match migration:
                case CreateEntity(entity=entity):
                    self.data.check_entity_not_exists(entity.name)
                    self.sql.execute(sql)
                    if entity.layout.type == LayoutType.HNHM:
                        attributes = {}
                        groups = {}
                    else:
                        attributes = entity.attributes
                        groups = entity.groups
                    self.data.entities[entity.name] = Entity(
                        name=entity.name,
                        layout=entity.layout,
                        doc=entity.doc,
                        keys=entity.keys,
                        attributes=attributes,
                        groups=groups,
                    )

                case RemoveEntity(entity=entity):
                    self.data.check_entity_exists(entity.name)
                    self.sql.execute(sql)
                    del self.data.entities[entity.name]

                case RecreateEntityView(entity=entity):
                    self.sql.execute(sql)
                    self.data.entities_views.add(entity.name)

                case RemoveEntityView(entity=entity):
                    if entity.name not in self.data.entities_views:
                        raise HnhmError(f"Entity's View '{entity.name}' doesn't exist.")
                    self.sql.execute(sql)
                    self.data.entities_views.remove(entity.name)

                case CreateAttribute(entity=entity, attribute=attribute):
                    self.data.check_attribute_not_exists(entity.name, attribute.name)
                    self.sql.execute(sql)
                    self.data.entities[entity.name].attributes[attribute.name] = attribute

                case RemoveAttribute(entity=entity, attribute=attribute):
                    self.data.check_attribute_exists(entity.name, attribute.name)
                    self.sql.execute(sql)
                    del self.data.entities[entity.name].attributes[attribute.name]

                case CreateGroup(entity=entity, group=group):
                    self.data.check_group_not_exists(entity.name, group.name)
                    self.sql.execute(sql)
                    self.data.entities[entity.name].groups[group.name] = group

                case RemoveGroup(entity=entity, group=group):
                    self.data.check_group_exists(entity.name, group.name)
                    self.sql.execute(sql)
                    del self.data.entities[entity.name].groups[group.name]

                case AddGroupAttribute(entity=entity, group=group, attribute=attribute):
                    self.data.check_group_attribute_not_exists(
                        entity.name, group.name, attribute.name
                    )
                    self.sql.execute(sql)
                    self.data.entities[entity.name].groups[group.name].attributes[
                        attribute.name
                    ] = attribute

                case RemoveGroupAttribute(
                    entity=entity, group=group, attribute=attribute
                ):
                    self.data.check_group_attribute_exists(
                        entity.name, group.name, attribute.name
                    )
                    self.sql.execute(sql)
                    del (
                        self.data.entities[entity.name]
                        .groups[group.name]
                        .attributes[attribute.name]
                    )

                case CreateLink(link=link):
                    self.data.check_link_not_exists(link.name)
                    self.sql.execute(sql)
                    self.data.links[link.name] = link

                case RemoveLink(link=link):
                    self.data.check_link_exists(link.name)
                    self.sql.execute(sql)
                    del self.data.links[link.name]

                case _:
                    raise HnhmError(f"Unknown migration: '{migration}'")



import pydantic

from .group import Group
from .layout import Layout
from .attribute import Attribute


class Entity(pydantic.BaseModel):
    """Entity core representation."""

    name: str
    layout: Layout
    doc: str
    keys: list[Attribute]
    attributes: dict[str, Attribute]
    groups: dict[str, Group]

    @property
    def sk(self) -> str:
        return f"{self.name}_sk"

    def __str__(self):
        return f"<Entity '{self.name}' layout='{self.layout}'>"

    def __hash__(self):
        return hash(f"{self.name}_{self.layout.type}")

    def __eq__(self, other):
        return hash(self) == hash(other)



from pathlib import Path

from hnhm.core import Storage, HnhmStorageData


class FileStorage(Storage):
    def __init__(self, file_name: str):
        self.file_name = file_name

    def load(self) -> HnhmStorageData:
        if Path(self.file_name).is_file():
            return HnhmStorageData.parse_file(self.file_name)
        else:
            return HnhmStorageData(entities={}, entities_views=set(), links={})

    def save(self, data: HnhmStorageData):
        with open(self.file_name, "w") as f:
            f.write(data.json(ensure_ascii=False, indent=2))



import abc
import inspect

from hnhm.core import Link, Layout, HnhmError, LayoutType, LinkElement

from .hnhm_entity import HnhmEntity


class HnhmLinkElement:
    def __init__(self, *, entity: HnhmEntity, comment: str):
        self.entity = entity
        self.comment = comment

    def to_core(self) -> LinkElement:
        entity = self.entity.to_core()
        if entity.layout.type != LayoutType.HNHM:
            raise HnhmError(
                f"Link doesn't support entities with LayoutType='{entity.layout.type}'."
                " Please, use elements with HNHM layout type."
            )

        if not self.comment:
            raise HnhmError(
                f"Passed empty comment='{self.comment}' entity='{entity.name}'."
                " Please, provide a comment for each LinkElement."
            )

        return LinkElement(entity=entity, comment=self.comment)


class HnhmLink(abc.ABC):
    __layout__: Layout
    __keys__: list[HnhmEntity]

    def to_core(self) -> Link:
        inspected = dict(inspect.getmembers(self))

        if not inspected.get("__layout__"):
            raise HnhmError(
                f"Layout not found for link: '{self}'."
                " Please, specify Layout via '__layout__' attribute."
            )
        layout: Layout = inspected["__layout__"]

        name = layout.name

        if not inspected.get("__doc__"):
            raise HnhmError(
                f"Doc not found or empty for link: '{name}'."
                " Please, write a documentation for your link."
            )
        doc: str = inspected["__doc__"]

        if not inspected.get("__keys__"):
            raise HnhmError(
                f"At least one Key is required for link='{name}'."
                " Please, specify link's keys via the '__keys__' attribute."
            )
        keys_hnhm: list[HnhmEntity] = inspected["__keys__"]
        keys = [key.to_core() for key in keys_hnhm]

        elements = []
        for class_attribute in inspected.values():
            if not isinstance(class_attribute, HnhmLinkElement):
                continue
            link_element = class_attribute.to_core()
            elements.append(link_element)

        if len(elements) < 2:
            raise HnhmError(
                f"At least two LinkElements are required for link='{name}'."
                " Please, specify more than one elements."
            )

        return Link(
            name=name,
            layout=layout,
            doc=doc,
            elements=elements,
            keys=keys,
        )



import abc

from .tasks import Task
from .migrations import Migration


class Sql(abc.ABC):
    """Generates and executes sql code."""

    @abc.abstractmethod
    def generate_sql(self, migration_or_task: Migration | Task) -> str:
        raise NotImplementedError

    @abc.abstractmethod
    def execute(self, sql: str):
        raise NotImplementedError


class FakeSql(Sql):
    def generate_sql(self, migration_or_task: Migration | Task) -> str:
        return ""

    def execute(self, sql: str):
        pass



from .flow import Flow
from .hnhm import HnHm
from .file_storage import FileStorage
from .hnhm_registry import HnhmRegistry
from .postgres.sql import PostgresSqlalchemySql
from .hnhm_link import HnhmLink, HnhmLinkElement
from .core import FakeSql, HnhmError, InMemStorage
from .hnhm_entity import Layout, HnhmEntity, LayoutType
from .hnhm_attribute import String, Integer, Timestamp, ChangeType

__all__ = [
    "String",
    "Integer",
    "Timestamp",
    "ChangeType",
    "HnhmEntity",
    "HnhmError",
    "Layout",
    "FakeSql",
    "LayoutType",
    "HnHm",
    "PostgresSqlalchemySql",
    "InMemStorage",
    "FileStorage",
    "Flow",
    "HnhmLink",
    "HnhmLinkElement",
    "HnhmRegistry",
]



import pydantic

from .attribute import Attribute, ChangeType


class Group(pydantic.BaseModel):
    """Group core representation."""

    name: str
    entity_name: str
    attributes: dict[str, Attribute]
    change_type: ChangeType

    @property
    def table(self):
        return f"group__{self.entity_name}__{self.name}"

    def __str__(self):
        return f"<Group '{self.name}' change_type='{self.change_type}'>"

    def __hash__(self):
        return hash(self.table)

    def __eq__(self, other):
        return hash(self) == hash(other)



from textwrap import dedent

import jinja2
from sqlalchemy import create_engine
from sqlalchemy.engine import URL, Engine

from hnhm.core import (
    Sql,
    Task,
    Type,
    LoadHub,
    LoadLink,
    HnhmError,
    LoadGroup,
    Migration,
    ChangeType,
    CreateLink,
    LayoutType,
    RemoveLink,
    CreateGroup,
    RemoveGroup,
    CreateEntity,
    RemoveEntity,
    LoadAttribute,
    CreateAttribute,
    RemoveAttribute,
    RemoveEntityView,
    AddGroupAttribute,
    RecreateEntityView,
    RemoveGroupAttribute,
)

PG_TYPES = {
    Type.STRING: "TEXT",
    Type.INTEGER: "INTEGER",
    Type.TIMESTAMP: "TIMESTAMPTZ",
}


def generate_sql(migration_or_task: Migration | Task, jinja: jinja2.Environment) -> str:
    match migration_or_task:
        case CreateEntity(entity=entity):
            if entity.layout.type == LayoutType.HNHM:
                template = jinja.get_template("create_hub.sql")
                columns = []
                columns_types = []
                for key in entity.keys:
                    columns.append(key.name)
                    column_type = PG_TYPES[key.type]
                    columns_types.append(f"{column_type} NOT NULL")

            elif entity.layout.type == LayoutType.STAGE:
                template = jinja.get_template("create_stage.sql")
                columns = []
                columns_types = []
                for attribute in entity.attributes.values():
                    columns.append(attribute.name)
                    columns_types.append(PG_TYPES[attribute.type])

            else:
                raise HnhmError(f"Unknown LayoutType='{entity.layout.type}'")

            return template.render(
                name=entity.name,
                columns=columns,
                columns_types=columns_types,
            )

        case RecreateEntityView(entity=entity):
            template = jinja.get_template("update_entity_view.sql")

            view_name = f"entity__{entity.name}"
            sk = entity.sk
            hub = f"hub__{entity.name}"

            selects = []
            for attribute in entity.attributes.values():
                selects.append((attribute.table, attribute.name, attribute.name))

            for group in entity.groups.values():
                for attribute in group.attributes.values():
                    selects.append(
                        (group.table, attribute.name, f"{group.name}__{attribute.name}")
                    )

            return template.render(
                view_name=view_name,
                sk=sk,
                hub=hub,
                selects=selects,
                attributes=entity.attributes.values(),
                groups=entity.groups.values(),
            )

        case CreateAttribute(entity=entity, attribute=attribute):
            attribute_type = PG_TYPES[attribute.type]

            if entity.layout.type == LayoutType.STAGE:
                return f"ALTER TABLE stg__{entity.name} ADD COLUMN {attribute.name} {attribute_type}"

            time_columns = ["valid_from TIMESTAMPTZ NOT NULL"]
            if attribute.change_type == ChangeType.NEW:
                time_columns.append("valid_to TIMESTAMPTZ")

            template = jinja.get_template("create_attribute.sql")
            return template.render(
                entity_name=entity.name,
                attribute_name=attribute.name,
                attribute_type=attribute_type,
                time_columns=time_columns,
            )

        case CreateGroup(entity=entity, group=group):
            columns = []
            for attribute in group.attributes.values():
                column_type = PG_TYPES[attribute.type]
                columns.append(f"{attribute.name} {column_type}")

            time_columns = ["valid_from TIMESTAMPTZ NOT NULL"]
            if group.change_type == ChangeType.NEW:
                time_columns.append("valid_to TIMESTAMPTZ")

            template = jinja.get_template("create_group.sql")
            return template.render(
                entity_name=entity.name,
                group_name=group.name,
                columns=columns,
                time_columns=time_columns,
            )

        case AddGroupAttribute(entity=entity, group=group, attribute=attribute):
            attribute_type = PG_TYPES[attribute.type]
            return f"ALTER TABLE group__{entity.name}__{group.name} ADD COLUMN {attribute.name} {attribute_type}"

        case CreateLink(link=link):
            entities = []
            for link_element in link.elements:
                entities.append(link_element.entity.name)

            primary_keys = ["valid_from"]
            for key in link.keys:
                primary_keys.append(f"{key.entity.name}_sk")

            template = jinja.get_template("create_link.sql")
            return template.render(
                name=link.name,
                entities=entities,
                primary_keys=primary_keys,
            )

        case RemoveEntity(entity=entity):
            if entity.layout.type == LayoutType.HNHM:
                table_name = f"hub__{entity.name}"
            else:
                table_name = f"stg__{entity.name}"
            return f"DROP TABLE {table_name}"

        case RemoveAttribute(entity=entity, attribute=attribute):
            if entity.layout.type == LayoutType.STAGE:
                return f"ALTER TABLE stg__{entity.name} DROP COLUMN {attribute.name}"

            return f"DROP TABLE attr__{entity.name}__{attribute.name}"

        case RemoveGroup(entity=entity, group=group):
            return f"DROP TABLE group__{entity.name}__{group.name}"

        case RemoveGroupAttribute(entity=entity, group=group, attribute=attribute):
            return f"ALTER TABLE group__{entity.name}__{group.name} DROP COLUMN {attribute.name}"

        case RemoveLink(link=link):
            return f"DROP TABLE link__{link.name}"

        case RemoveEntityView(entity=entity):
            view_name = f"entity__{entity.name}"
            return f"DROP VIEW {view_name}"

        case LoadHub(
            source=source,
            target=target,
            business_time_field=business_time_field,
            keys_mapping=keys_mapping,
        ):
            source_keys = [key_source.name for key_source in keys_mapping.values()]
            source_sk_components = (f"{key}::TEXT" for key in source_keys)
            source_sk_components = "|| '-' ||".join(source_sk_components)
            source_sk = f"MD5({source_sk_components})"

            target_keys = [key.name for key in target.keys]

            template = jinja.get_template("load_hub.sql")
            return template.render(
                source_name=source.name,
                source_sk=source_sk,
                source_keys=source_keys,
                target_name=target.name,
                target_keys=target_keys,
                business_time_field=business_time_field.name,
            )

        case LoadAttribute(
            source=source,
            target=target,
            business_time_field=business_time_field,
            keys_mapping=keys_mapping,
            source_attribute=source_attribute,
            target_attribute=target_attribute,
        ):
            source_table = f"stg__{source.name}"

            source_keys = [key_source.name for key_source in keys_mapping.values()]
            source_sk_components = (f"{key}::TEXT" for key in source_keys)
            source_sk_components = "|| '-' ||".join(source_sk_components)
            source_sk = f"MD5({source_sk_components})"

            source_attributes = [source_attribute.name]

            target_table = f"attr__{target.name}__{target_attribute.name}"
            target_sk = f"{target.name}_sk"
            target_sks = [f"{target.name}_sk"]

            target_attributes = [target_attribute.name]

            if target_attribute.change_type == ChangeType.IGNORE:
                template = jinja.get_template("load_ignore.sql")
            elif target_attribute.change_type == ChangeType.UPDATE:
                template = jinja.get_template("load_update.sql")
            elif target_attribute.change_type == ChangeType.NEW:
                template = jinja.get_template("load_new.sql")
            else:
                raise HnhmError(f"Unknown change type: '{target_attribute.change_type}'.")

            return template.render(
                source_table=source_table,
                source_sk=source_sk,
                source_sks=[source_sk],
                source_attributes=source_attributes,
                target_table=target_table,
                target_sk=target_sk,
                target_sks=target_sks,
                target_attributes=target_attributes,
                business_time_field=business_time_field.name,
                extra_sks=[],
            )

        case LoadGroup(
            source=source,
            target=target,
            group=group,
            business_time_field=business_time_field,
            keys_mapping=keys_mapping,
            attributes_mapping=attributes_mapping,
        ):
            source_table = f"stg__{source.name}"

            source_keys = [key_source.name for key_source in keys_mapping.values()]
            source_sk_components = (f"{key}::TEXT" for key in source_keys)
            source_sk_components = "|| '-' ||".join(source_sk_components)
            source_sk = f"MD5({source_sk_components})"

            source_attributes = [
                source_attribute.name for source_attribute in attributes_mapping.values()
            ]

            target_table = f"group__{target.name}__{group.name}"
            target_sk = f"{target.name}_sk"
            target_sks = [target_sk]

            target_attributes = [
                target_attribute.name for target_attribute in attributes_mapping
            ]

            if group.change_type == ChangeType.IGNORE:
                template = jinja.get_template("load_ignore.sql")
            elif group.change_type == ChangeType.UPDATE:
                template = jinja.get_template("load_update.sql")
            elif group.change_type == ChangeType.NEW:
                template = jinja.get_template("load_new.sql")
            else:
                raise HnhmError(f"Unknown change type: '{group.change_type}'.")

            return template.render(
                source_table=source_table,
                source_sk=source_sk,
                source_sks=[source_sk],
                source_attributes=source_attributes,
                target_table=target_table,
                target_sk=target_sk,
                target_sks=target_sks,
                target_attributes=target_attributes,
                business_time_field=business_time_field.name,
                extra_sks=[],
            )

        case LoadLink(
            source=source,
            link=link,
            business_time_field=business_time_field,
            keys_mapping=keys_mapping,
        ):
            source_table = f"stg__{source.name}"

            target_sks = []
            source_sks = []
            for entity_name, entity_keys_mapping in keys_mapping.items():
                source_keys = [
                    key_source.name for key_source in entity_keys_mapping.values()
                ]
                source_sk_components = (f"{key}::TEXT" for key in source_keys)
                source_sk_components = "|| '-' ||".join(source_sk_components)
                source_sk = f"MD5({source_sk_components})"

                target_sks.append(f"{entity_name}_sk")
                source_sks.append(source_sk)

            target_table = f"link__{link.name}"

            template = jinja.get_template("load_new.sql")
            return template.render(
                source_table=source_table,
                source_sks=source_sks,
                source_attributes=[],
                target_table=target_table,
                target_sks=target_sks,
                target_attributes=[],
                business_time_field=business_time_field.name,
                extra_sks=["valid_from"],
            )

        case _:
            migration_or_task_type = type(migration_or_task)
            raise HnhmError(
                f"Unknown migration or task: '{migration_or_task}' with type='{migration_or_task_type}'."
            )


class PostgresSqlalchemySql(Sql):
    def __init__(
        self,
        *,
        host: str = "localhost",
        port: int = 5432,
        database: str = "postgres",
        user: str = "postgres",
        password: str | None = None,
        engine: Engine | None = None,
    ):
        if engine:
            self.engine = engine
        else:
            connection_url = URL.create(
                "postgresql+psycopg2",
                username=user,
                password=password,
                host=host,
                database=database,
                port=port,
            )
            self.engine = create_engine(connection_url)

        self.jinja = jinja2.Environment(
            loader=jinja2.PackageLoader("hnhm.postgres", "sql_templates")
        )
        self.jinja.globals.update(zip=zip)

    @classmethod
    def with_engine(cls, engine: Engine):
        if engine.url.drivername != "postgresql+psycopg2":
            raise HnhmError(
                f"Wrong driver name '{engine.url.drivername}'. Required: 'postgresql+psycopg2'."
            )
        return cls(engine=engine)

    def generate_sql(self, migration_or_task: Migration | Task) -> str:
        return generate_sql(migration_or_task, self.jinja)

    def execute(self, sql: str, debug: bool = False):
        if debug:
            print(dedent(sql).strip())

        conn = None
        try:
            conn = self.engine.connect()
            conn = conn.execution_options(isolation_level="AUTOCOMMIT")
            conn.execute(sql)
        except Exception as e:
            raise e
        finally:
            if conn:
                conn.close()
            self.engine.dispose()



from hnhm.core import (
    Task,
    Entity,
    LoadHub,
    LoadLink,
    Attribute,
    HnhmError,
    LoadGroup,
    LoadAttribute,
)

from .hnhm_link import HnhmLink
from .hnhm_attribute import HnhmAttribute
from .hnhm_entity import HnhmEntity, LayoutType


class Flow:
    def __init__(self, *, source: HnhmEntity, business_time_field: HnhmAttribute):
        source = source.to_core()
        if source.layout.type != LayoutType.STAGE:
            raise HnhmError(
                f"Flow doesn't support loading from entities with layout type '{source.layout.type}'."
                " Please, use Flow to load data only from STAGE entities."
            )

        self.source = source
        self.business_time_field = business_time_field.to_core(source.name)

        # {entity_name -> {attribute_target -> attribute_source}}
        self._entities_keys_mappings: dict[str, dict[Attribute, Attribute]] = {}

        # {entity_name -> target}
        self._hubs: dict[str, Entity] = {}

        # {entity_name -> {attribute_target -> attribute_source}}
        self._attributes: dict[str, dict[Attribute, Attribute]] = {}

        # {entity_name -> {group_name -> {attribute_target -> attribute_source}}}
        self._groups: dict[str, dict[str, dict[Attribute, Attribute]]] = {}

        # {link_name -> {entity_name -> {attribute_target -> attribute_source}}}
        self._links = {}

    @property
    def tasks(self) -> list[Task]:
        tasks = []
        hub_tasks = {}
        for target_name, target in self._hubs.items():
            keys_mapping = self._entities_keys_mappings[target_name]
            hub_tasks[target_name] = LoadHub(
                source=self.source,
                target=target,
                keys_mapping=keys_mapping,
                business_time_field=self.business_time_field,
            )

        for target_name, attributes_mapping in self._attributes.items():
            target = self._hubs[target_name]
            depends_on = [hub_tasks[target_name].id]
            keys_mapping = self._entities_keys_mappings[target_name]

            for target_attribute, source_attribute in attributes_mapping.items():
                tasks.append(
                    LoadAttribute(
                        source=self.source,
                        target=target,
                        keys_mapping=keys_mapping,
                        source_attribute=source_attribute,
                        target_attribute=target_attribute,
                        business_time_field=self.business_time_field,
                        depends_on=depends_on,
                    )
                )

        for target_name, groups_mapping in self._groups.items():
            target = self._hubs[target_name]
            depends_on = [hub_tasks[target_name].id]
            keys_mapping = self._entities_keys_mappings[target_name]
            for group_name, attributes_mapping in groups_mapping.items():
                group = target.groups[group_name]
                tasks.append(
                    LoadGroup(
                        source=self.source,
                        target=target,
                        group=group,
                        keys_mapping=keys_mapping,
                        attributes_mapping=attributes_mapping,
                        business_time_field=self.business_time_field,
                        depends_on=depends_on,
                    )
                )

        for link_name, link in self._links.items():
            keys_mapping = {}
            depends_on = []
            for link_element in link.elements:
                entity_name = link_element.entity.name
                depends_on.append(hub_tasks[entity_name].id)
                keys_mapping[entity_name] = self._entities_keys_mappings[entity_name]
            tasks.append(
                LoadLink(
                    source=self.source,
                    link=link,
                    keys_mapping=keys_mapping,
                    business_time_field=self.business_time_field,
                    depends_on=depends_on,
                )
            )
        tasks.extend(hub_tasks.values())
        tasks = sorted(tasks, key=lambda t: t.priority)
        return tasks

    def load(
        self,
        target: HnhmEntity | HnhmLink,
        mapping: dict[HnhmAttribute, HnhmAttribute] | None = None,
    ):
        match target:
            case HnhmEntity():
                target_entity = target.to_core()

                if not mapping:
                    raise HnhmError(
                        f"Mapping is required to load data for entity='{target_entity.name}'."
                        " Please, provide attributes mapping."
                    )

                if target_entity.layout.type != LayoutType.HNHM:
                    raise HnhmError(
                        "Loading is only supported to the entities with the LayoutType='HNHM'."
                        f" Your target entity='{target_entity.name}' has LayoutType='{target_entity.layout.type}'."
                    )

                keys_mapping = {}
                groups_mapping = {}
                attributes_mapping = {}
                for attribute_target, attribute_source in mapping.items():
                    attribute_target = attribute_target.to_core(target_entity.name)
                    attribute_source = attribute_source.to_core(self.source.name)
                    if attribute_target in target_entity.keys:
                        keys_mapping[attribute_target] = attribute_source
                    elif attribute_target.group:
                        group_name = attribute_target.group
                        if group_name not in groups_mapping:
                            groups_mapping[group_name] = {}
                        groups_mapping[group_name][attribute_target] = attribute_source
                    else:
                        attributes_mapping[attribute_target] = attribute_source

                # Check all attributes for a Group were mapped
                for group_name, group in target_entity.groups.items():
                    group_mapping = groups_mapping.get(group_name)
                    if not group_mapping:
                        continue

                    for attribute in group.attributes.values():
                        if attribute not in group_mapping:
                            attribute_full_name = (
                                f"{target_entity.name}.{group_name}.{attribute.name}"
                            )
                            raise HnhmError(
                                f"Mapping not found for the attribute '{attribute_full_name}'."
                                f" Please, provide all mappings for the group: '{target_entity.name}.{group_name}'."
                            )

                missing_keys = set(target_entity.keys) - set(keys_mapping.keys())
                if missing_keys:
                    missing_keys_names = [key.name for key in missing_keys]
                    raise HnhmError(
                        f"Found missing mappings for entity='{target_entity.name}' keys."
                        f" Missing mappings for keys: {missing_keys_names}."
                    )

                if target_entity.name in self._entities_keys_mappings:
                    raise HnhmError(
                        f"Found duplicated load for entity='{target_entity.name}'."
                        " Please, use single load() block for single entity."
                    )

                self._entities_keys_mappings[target_entity.name] = keys_mapping
                self._hubs[target_entity.name] = target_entity
                self._attributes[target_entity.name] = attributes_mapping
                self._groups[target_entity.name] = groups_mapping

            case HnhmLink():
                link = target.to_core()

                for element in link.elements:
                    if element.entity.name not in self._entities_keys_mappings:
                        raise HnhmError(
                            f"Keys mapping not found for entity='{element.entity.name}'."
                            " Please, provide mapping the entity keys."
                        )

                if link.name in self._links:
                    raise HnhmError(
                        f"Found duplicated load for link='{link.name}'."
                        " Please, use single load() block for single link."
                    )

                self._links[link.name] = link

            case _:
                target_type = type(target)
                raise HnhmError(f"Unknown target type '{target_type}'.")

        return self
